indexing

	description: "Base class for simple CGI programs. Outputs XHTML."

	usage: "Never write to stdout, use `puts' or `add_data' instead."
	standards: "Based on CGI spec 1.1, draft-coar-cgi-v11-03."

	known_bugs:
		"1. Does not handle inputs with same name (checkboxes) correctly."

	todo:
		"1. should use binary search allowed_new_line_tag"

	author: "Berend de Boer"
	date: "$Date: 2004/12/18 $"
	revision: "$Revision: #12 $"


deferred class

	EPX_CGI


inherit

	STDC_CURRENT_PROCESS

	EPX_XHTML_WRITER
		rename
			make as make_xhtml_writer
		end

	EPX_FILE_SYSTEM
		export
			{NONE} all
		end

	EPX_MIME_FIELD_NAMES
		export
			{NONE} all
		end

	EPX_HTTP_METHODS
		export
			{NONE} all
		end


feature {NONE} -- Initialization

	make is
		do
			make_no_rescue
		rescue
			handle_fatal_error
		end

	make_no_rescue is
		do
			make_xhtml_writer
			debug ("cgi")
				dump_input
				exit_with_failure
			end
			-- Output is accumulated in the XML string. By calling
			-- `stdout.write_string' output is written now. That's only done
			-- for the header.
			stdout.set_full_buffering (default_pointer, 8192)
			execute
			-- Write output now.
			stdout.write_string (as_uc_string)
		end


feature -- Output

	execute is
			-- To be implemented by child.
		deferred
		end


feature {NONE} -- When script fails

	handle_fatal_error is
			-- Your last change to emit something when the script has
			-- failed unexpectedly. Call only in rescue clause.
		require
			is_exception: exceptions.exception /= 0
		do
			if not is_header_written then
				content_text_plain
			end
			stdout.put_string ("A fatal error prevents this program from continuing.%N")
			if exceptions.is_developer_exception then
				stdout.put_string ("The error message was: ")
				stdout.put_string (exceptions.developer_exception_name)
				stdout.put_string ("%N")
			end
			stdout.put_string ("Exception code: ")
			stdout.put_string (exceptions.exception.out)
			stdout.put_string ("%N")
			stdout.flush
			exit_with_failure
		end


feature -- Debug support

	dump_input is
			-- Write cgi input to $TMPDIR/cgi_input.
			-- First line contains the content header, is not actually in input!
		local
			file: STDC_BINARY_FILE
			clen: INTEGER
			i: INTEGER
		do
			create file.create_write (temporary_directory + "/cgi_input")
			if request_method.is_equal (http_method_GET) then
				file.puts (query_string)
			elseif request_method.is_equal (http_method_POST) then
				file.write_string (mime_header)
				from
					i := 0
					clen := content_length
				until
					i >= clen
				loop
					stdin.get_character
					file.putc (stdin.last_byte)
					i := i + 1
				end
			end
			file.close
		end


feature -- Standard variables

	auth_type: STRING is
			-- type of authentication used
		do
			Result := env_auth_type.value
		end

	content_type: STRING is
			-- MIME type of data when invoked with POST method
		do
			Result := env_content_type.value
		end

	content_length: INTEGER is
			-- length, in bytes, of data when invoked with POST method
		do
			if env_content_length.value.is_empty then
				Result := 0
			else
				Result := env_content_length.value.to_integer
			end
		end

	gateway_interface: STRING is
			-- Name and version of the gateway, for example CGI/1.1
		do
			Result := env_gateway_interface.value
		end

	http_accept: STRING is
			-- Contents of the Accept header line sent by the client
		do
			Result := env_http_accept.value
		end

	http_cookie: STRING is
			-- All cookies sent by the client in the form of key=value,
			-- semi-colon separated.
		do
			Result := env_http_cookie.value
		ensure
			http_cookie_not_void: Result /= Void
		end

	http_referer: STRING is
			-- Contents of the Referer header line.
		do
			Result := env_http_referer.value
		end

	http_user_agent: STRING is
			-- Name of the client program that is making the request.
		do
			Result := env_http_user_agent.value
		end

	path_info: STRING is
			-- Extra path information as it was passed to the server in
			-- the query URL
		do
			Result := env_path_info.value
		end

	path_translated: STRING is
			-- Extra path information translated to a final, usable
			-- form. The Web document root is prepended to the query
			-- path, and any other path translations are executed.
		do
			Result := env_path_translated.value
		end

	query_string: STRING is
			-- The input when invoked with the GET method.
		do
			Result := env_query_string.value
		ensure
			query_string_not_void: Result /= Void
		end

	remote_addr, remote_address: STRING is
			-- IP address of the client that made the request
		do
			Result := env_remote_addr.value
		end

	remote_host: STRING is
			-- name of the remote computer that made the request
		do
			Result := env_remote_host.value
		end

	remote_ident: STRING is
			-- user name as given by the ident protocol
		do
			Result := env_remote_ident.value
		end

	remote_user: STRING is
			-- name of the remote user that made the request
		do
			Result := env_remote_user.value
		end

	request_method: STRING is
			-- name of the method used to invoke the CGI
			-- application. Valid values are GET and POST
		do
			Result := env_request_method.value
		end

	script_name: STRING is
			-- name of script that was invoked
		do
			Result := env_script_name.value
		end

	server_name: STRING is
			-- domain name of the computer that is running the server software
		do
			Result := env_server_name.value
		end

	server_port: INTEGER is
			-- TCP port number on which the server that invoked the CGI
			-- application is operating
		local
			s: STRING
		do
			s := env_server_port.value
			if s.is_empty then
				Result := 0
			else
				Result := s.to_integer
			end
		end

	server_protocol: STRING is
			-- name of the protocol that the server is using and the
			-- version of that protocol. The protocol name and version
			-- are separated by a forward slash with no spaces, for
			-- instance HTTP/1.0
		do
			Result := env_server_protocol.value
		end

	server_software: STRING is
			-- name of the server that is handling the request
		do
			Result := env_server_software.value
		end


feature {NONE} -- Environment variables

	env_auth_type: STDC_ENV_VAR is
		once
			create Result.make ("AUTH_TYPE")
		end

	env_content_length: STDC_ENV_VAR is
		once
			create Result.make ("CONTENT_LENGTH")
		end

	env_content_type: STDC_ENV_VAR is
		once
			create Result.make ("CONTENT_TYPE")
		end

	env_gateway_interface: STDC_ENV_VAR is
		once
			create Result.make ("GATEWAY_INTERFACE")
		end

	env_http_accept: STDC_ENV_VAR is
		once
			create Result.make ("HTTP_ACCEPT")
		end

	env_http_cookie: STDC_ENV_VAR is
		once
			create Result.make ("HTTP_COOKIE")
		ensure
			env_http_cookie_not_void: Result /= Void
		end

	env_http_referer: STDC_ENV_VAR is
		once
			create Result.make ("HTTP_REFERER")
		end

	env_http_user_agent: STDC_ENV_VAR is
		once
			create Result.make ("HTTP_USER_AGENT")
		end

	env_path_info: STDC_ENV_VAR is
		once
			create Result.make ("PATH_INFO")
		end

	env_path_translated: STDC_ENV_VAR is
		once
			create Result.make ("PATH_TRANSLATED")
		end

	env_query_string: STDC_ENV_VAR is
		once
			create Result.make ("QUERY_STRING")
		end

	env_remote_addr: STDC_ENV_VAR is
		once
			create Result.make ("REMOTE_ADDR")
		end

	env_remote_host: STDC_ENV_VAR is
		once
			create Result.make ("REMOTE_HOST")
		end

	env_remote_user: STDC_ENV_VAR is
		once
			create Result.make ("REMOTE_USER")
		end

	env_remote_ident: STDC_ENV_VAR is
		once
			create Result.make ("REMOTE_IDENT")
		end

	env_request_method: STDC_ENV_VAR is
		once
			create Result.make ("REQUEST_METHOD")
		end

	env_script_name: STDC_ENV_VAR is
		once
			create Result.make ("SCRIPT_NAME")
		end

	env_server_name: STDC_ENV_VAR is
		once
			create Result.make ("SERVER_NAME")
		end

	env_server_port: STDC_ENV_VAR is
		once
			create Result.make ("SERVER_PORT")
		end

	env_server_protocol: STDC_ENV_VAR is
		once
			create Result.make ("SERVER_PROTOCOL")
		end

	env_server_software: STDC_ENV_VAR is
		once
			create Result.make ("SERVER_SOFTWARE")
		end


feature -- CGI headers

	content_text_html is
		do
			stdout.write_string ("Content-Type: text/html%N")
			finish_header
		end

	content_text_plain is
		do
			stdout.write_string ("Content-Type: text/plain%N")
			finish_header
		end

	finish_header is
			-- Finish the header by emitting an empty line.
			-- If `cookies' have been set, they are written as well.
		local
			cookie: EPX_HTTP_COOKIE
		do
			from
				cookies.start
			until
				cookies.after
			loop
				cookie := cookies.item_for_iteration
				stdout.write_string ("Set-Cookie: ")
				stdout.write_string (cookie.key)
				stdout.write_string ("=")
				stdout.write_string (cookie.value)
				if cookie.domain /= Void then
					stdout.write_string ("; domain=")
					stdout.write_string (cookie.domain)
				end
				if cookie.path /= Void then
					stdout.write_string ("; path=")
					stdout.write_string (cookie.path)
				end
				if cookie.secure then
					stdout.write_string ("; secure")
				end
				if cookie.expires /= Void then
					stdout.write_string ("; expires=")
					stdout.write_string (cookie.expires.format ("%%a, %%d-%%b-%%Y %%H:%%M:%%S GMT"))
				end
				stdout.write_character ('%N')
				cookies.forth
			end
			stdout.write_character ('%N')
			stdout.flush
		end

	location (a_url: STRING) is
			-- Redirect to `a_url' by emitting a Location header.
			-- Note that `a_url' MUST BE a FULL url, a partial URL has
			-- semantics that you probably would find surprising.
		require
			a_url_not_empty: a_url /= Void and then not a_url.is_empty
		do
			stdout.write_string (field_name_location)
			stdout.write_string (": ")
			stdout.write_string (a_url)
			stdout.write_string ("%N%N")
			-- flush here seems to cause a SIGTERM. Not sure why Apache
			-- would send this. So you might want to catch/ignore this signal.
			stdout.flush
		end


feature -- Cookies

	cookies: DS_HASH_TABLE [EPX_HTTP_COOKIE, STRING] is
			-- Cookies that will be returned to the browser.
		once
			create Result.make (4)
		ensure
			cookies_not_void: cookies /= Void
		end

	set_cookie (a_cookie: EPX_HTTP_COOKIE) is
			-- Add a new cookie that will be send to the browser then
			-- `context_text_html' is called.
		require
			cookie_not_void: a_cookie /= Void
			cookie_not_added: not cookies.has (a_cookie.key)
		do
			cookies.force_last (a_cookie, a_cookie.key)
		ensure
			cookie_known: cookies.has (a_cookie.key)
		end


feature -- Server push, multipart header

	content_multipart_x_mixed_replace (boundary: STRING) is
			-- Initiate server push.
		require
			valid_boundary: boundary /= Void and then not boundary.is_empty
		do
			stdout.write_string ("Content-Type: multipart/x-mixed-replace;boundary=")
			stdout.write_string (boundary)
			my_boundary := boundary
			stdout.write_string ("%N%N")
			-- start first part immediately
			stdout.write_string (my_boundary)
			stdout.write_string ("%N")
			stdout.flush
		end

	content_next_part is
			-- Write boundary so next part of multipart msg can be written.
		require
			multipart: is_multipart_message
		do
			stdout.write_character ('%N')
			stdout.write_string (my_boundary)
			stdout.write_character ('%N')
			stdout.flush
		end

	content_multipart_end is
			-- Write boundary of multipart.
		require
			multipart: is_multipart_message
		do
			stdout.write_character ('%N')
			stdout.write_string (my_boundary)
			stdout.write_string ("--%N")
			stdout.flush
		end

	is_multipart_message: BOOLEAN is
			-- Are we writing server push, multipart output?
		do
			Result := my_boundary /= Void
		end


feature -- Form input

	has_input: BOOLEAN is
			-- Is input passed to cgi program?
		do
			Result := not request_method.is_empty
		end

	has_key (key: STRING): BOOLEAN is
			-- Is `key' passed as parameter/form-data?
		require
			key_not_empty: key /= Void and then not key.is_empty
		do
			Result := cgi_data.has (key)
		ensure
			definition: Result = cgi_data.has (key)
		end

	is_meta_char (c: CHARACTER): BOOLEAN is
			-- Is `c' a commonly used meta character?
		do
			Result :=
				(c.code = 0) or else
				meta_chars.has (c)
		end

	meta_chars: STRING is
			-- Commonly used meta characters.
		once
			-- BdB: Check if this list is complete...
			Result := "`[]<>|&`\%%"
		ensure
			meta_chars_not_empty: meta_chars /= Void and then not meta_chars.is_empty
		end

	raw_value (key: STRING): STRING is
			-- Returns value for key.
			-- if key does not exist, the empty string is returned.
		require
			key_not_empty: key /= Void and then not key.is_empty
		do
			search_key (key)
			if found_key then
				create Result.make_from_string (found_key_item.value)
			else
				Result := ""
			end
		ensure
			raw_value_not_void: Result /= Void
		end

	remove_meta_chars (s: STRING) is
			-- If `s' contains meta characters, they're removed.
		require
			s_not_void: s /= Void
		local
			i: INTEGER
		do
			from
				i := 1
			until
				i > s.count
			loop
				if is_meta_char (s.item (i)) then
					s.remove (i)
				else
					i := i + 1
				end
			end
		end

	value (key: STRING): STRING is
			-- As `raw_value' but meta characters are removed
		require
			key_not_empty: key /= Void and then not key.is_empty
		do
			Result := raw_value (key)
			remove_meta_chars (Result)
		ensure
			value_not_void: Result /= Void
		end


feature {NONE} -- Cached key/value

	assert_key_value_pairs_created is
			-- Make sure passed form contents (if any) is parsed.
			-- Cookies are added as key/value pairs as well.
			-- If there is form data with the same name, it overrides the
			-- cookies value (yes, that's a feature).
		local
			given_cookies: DS_HASH_TABLE [EPX_KEY_VALUE, STRING]
			kv: EPX_KEY_VALUE
		do
			if cgi_data = Void then
				if has_input then
					fill_key_value_pairs
				else
					create cgi_data.make_default
				end
				if not http_cookie.is_empty then
					given_cookies := url_encoder.do_url_encoded_to_field_name_value_pair (http_cookie, ';')
					from
						given_cookies.start
					until
						given_cookies.after
					loop
						kv := given_cookies.item_for_iteration
						if not cgi_data.has (kv.key) then
							cgi_data.force (kv, kv.key)
						end
						given_cookies.forth
					end
				end
			end
		ensure
			cgi_data_not_void: cgi_data /= Void
		end

	cgi_data: DS_HASH_TABLE [EPX_KEY_VALUE, STRING]
			-- Array with key-value pairs;
			-- If you want to loop through the values, make sure to call
			-- `assert_key_value_pairs_created'.

	cgi_input: STRING
			-- Raw input data.

	fill_key_value_pairs is
			-- Fill `cgi_data', recognizes default and multipart encodings.
		require
			key_value_pair_not_set: cgi_data = Void
			there_is_input: has_input
		local
			ctype: STRING
		do
			create ctype.make_from_string (content_type)
			ctype.to_lower
			if ctype.is_empty or else ctype.is_equal (mime_type_application_x_www_form_urlencoded) then
				fill_key_value_pairs_from_standard_encoding
			elseif ctype.substring_index (mime_type_multipart_form_data, 1) = 1 then
				fill_key_value_pairs_from_multipart_encoding
			end
		ensure
			cgi_data_not_void: cgi_data /= Void
		end

	found_key: BOOLEAN
			-- Was `search_key' successful?

	found_key_item: EPX_KEY_VALUE
			-- Item found by `search_key'.

	search_key (key: STRING) is
			-- Locate key in key/value pairs.
			-- Sets `found_key' if found.
		require
			valid_key: key /= Void and then not key.is_empty
		do
			assert_key_value_pairs_created
			cgi_data.search (key)
			if cgi_data.found then
				found_key := True
				found_key_item := cgi_data.found_item
			else
				found_key := False
				found_key_item := Void
			end
		ensure
			found_implies_item: found_key = (found_key_item /= Void)
		end

	set_cgi_input is
			-- Set `cgi_input'. When method is POST, the variable
			-- Content-Length must be specified.
		require
			content_length_set: request_method.is_equal (http_method_POST) implies not env_content_length.value.is_empty
		local
			len: INTEGER
			i: INTEGER
		do
			if request_method.is_equal (http_method_GET) then
				cgi_input := query_string
			elseif request_method.is_equal (http_method_POST) then
				from
					len := content_length
					i := 0
					create cgi_input.make (len)
				until
					i = len
				loop
					stdin.read_character
					cgi_input.append_character (stdin.last_character)
					i := i + 1
				end
			end
		end


feature {NONE} -- Standard or multipart key/value filling

	fill_key_value_pairs_from_multipart_encoding is
			-- Posted data.
		local
			parser: EPX_MIME_PARSER
		do
			create parser.make_from_file (stdin)
			parser.set_header (mime_header)
			parser.parse
			if not parser.syntax_error then
				cgi_data := url_encoder.mime_encoded_to_field_name_value_pair (Void, parser.part.body)
			else
				exceptions.raise ("Syntax error detected during parsing of POSTed data.")
			end
		ensure
			cgi_data_not_void: cgi_data /= Void
		end

	fill_key_value_pairs_from_standard_encoding is
			-- Read query values from url request.
		do
			set_cgi_input
			if cgi_input /= Void then
				cgi_data := url_encoder.url_encoded_to_field_name_value_pair (cgi_input)
			end
		ensure
			cgi_data_not_void: cgi_data /= Void
		end

	mime_header: STRING is
			-- Suitable MIME header for POST data.
		local
			fixed_content_type: STRING
			i: INTEGER
			value_start,
			value_end: INTEGER
		once
			-- Opera contains a boundary parameter with invalid characters.
			-- Fix this by making sure the boundary parameter is properly quoted.
			create fixed_content_type.make_from_string (content_type)
			i := fixed_content_type.substring_index ("boundary=", 1)
			if i /= 0 then
				value_start := i + 9
				if value_start <= fixed_content_type.count then
					if fixed_content_type.item (value_start) /= '"' then
						STRING_.insert_character (fixed_content_type, '"', value_start)
						from
							value_end := value_start
						until
							value_end > fixed_content_type.count or else
							fixed_content_type.item (value_end) = ' ' or else
							fixed_content_type.item (value_end) = ';' or else
							fixed_content_type.item (value_end) = '%T' or else
							fixed_content_type.item (value_end) = '%N' or else
							fixed_content_type.item (value_end) = '%R'
						loop
							value_end := value_end + 1
						end
						if value_end > fixed_content_type.count then
							fixed_content_type.append_character ('"')
						else
							STRING_.insert_character (fixed_content_type, '"', value_end)
						end
					end
				end
			end
			create Result.make (field_name_content_type.count + 2 + fixed_content_type.count + 4)
			Result.append_string (field_name_content_type)
			Result.append_character (':')
			Result.append_character (' ')
			Result.append_string (fixed_content_type)
			Result.append_character ('%R')
			Result.append_character ('%N')
			Result.append_character ('%R')
			Result.append_character ('%N')
		ensure
			result_not_empty: Result /= Void and then not Result.is_empty
		end


feature {NONE} -- Implementation

	my_boundary: STRING
			-- Used for server push, multipart header.


	url_encoder: expanded EPX_URL_ENCODING
			-- Decoding of encoded passed values


end
